#!/usr/bin/env python

"""Tests for `arcos_py` package."""

from numpy import int64
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from arcos4py import ARCOS
from arcos4py.tools._errors import noDataError


@pytest.fixture
def no_bin_data():
    """
    pytest fixture to generate test data
    """
    data = [item for i in range(10) for item in list(range(1, 11))]
    m = [0 for i in range(100)]
    d = {'id': data, 'time': data, 'm': m, 'x': data}
    print(d)
    df = pd.DataFrame(d)
    return df


def test_empty_data(no_bin_data: pd.DataFrame):
    with pytest.raises(noDataError, match='Input is empty'):
        test_data = no_bin_data[no_bin_data['m'] > 0]
        pos = ['x']
        ts = ARCOS(
            test_data, posCols=pos, frame_column='time', id_column='id', measurement_column='m', clid_column='clTrackID'
        )
        ts.trackCollev(eps=1, minClsz=1, nPrev=2)


def test_1_central_1_prev():
    df_in = pd.read_csv('tests/testdata/1central_in.csv')
    df_true = pd.read_csv('tests/testdata/1central_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true, check_dtype=False)


def test_1_central_2_prev():
    df_in = pd.read_csv('tests/testdata/1central_in.csv')
    df_true = pd.read_csv('tests/testdata/1central2prev_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1, minClsz=1, nPrev=2)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true, check_dtype=False)


def test_1_central_3D():
    df_in = pd.read_csv('tests/testdata/1central3D_in.csv')
    df_true = pd.read_csv('tests/testdata/1central3D_res.csv')
    pos = ['x', 'y', 'z']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x', 'y', 'z'])
    assert_frame_equal(out, df_true)


def test_1_central_growing():
    df_in = pd.read_csv('tests/testdata/1centralGrowing_in.csv')
    df_true = pd.read_csv('tests/testdata/1centralGrowing_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_2_central_growing():
    df_in = pd.read_csv('tests/testdata/2centralGrowing_in.csv')
    df_true = pd.read_csv('tests/testdata/2centralGrowing_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_2_with_1_common_symmetric():
    df_in = pd.read_csv('tests/testdata/2with1commonSym_in.csv')
    df_true = pd.read_csv('tests/testdata/2with1commonSym_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_2_with_1_common_asymmetric():
    df_in = pd.read_csv('tests/testdata/2with1commonAsym_in.csv')
    df_true = pd.read_csv('tests/testdata/2with1commonAsym_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_3_spreading_1_prev():
    df_in = pd.read_csv('tests/testdata/3spreading_in.csv')
    df_true = pd.read_csv('tests/testdata/3spreading_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_3_spreading_2_prev():
    df_in = pd.read_csv('tests/testdata/3spreading_in.csv')
    df_true = pd.read_csv('tests/testdata/3spreading2prev_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=2)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_5_overlapping_1_prev():
    df_in = pd.read_csv('tests/testdata/5overlapping_in.csv')
    df_true = pd.read_csv('tests/testdata/5overlapping_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_5_overlapping_2_prev():
    df_in = pd.read_csv('tests/testdata/5overlapping_in.csv')
    df_true = pd.read_csv('tests/testdata/5overlapping2prev_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=2)
    out = out.drop(columns=['m', 'x'])
    assert_frame_equal(out, df_true)


def test_6_overlapping():
    df_in = pd.read_csv('tests/testdata/6overlapping_in.csv')
    df_true = pd.read_csv('tests/testdata/6overlapping_res.csv')
    pos = ['x']
    ts = ARCOS(
        df_in, posCols=pos, frame_column='time', id_column='trackID', measurement_column='m', clid_column='clTrackID'
    )
    ts.bin_col = 'm'
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['m', 'x'])
    out['trackID'] = out['trackID'].astype(int64)
    assert_frame_equal(out, df_true)


def test_split_from_single():
    df_in = pd.read_csv('tests/testdata/1objSplit_in.csv')
    df_true = pd.read_csv('tests/testdata/1objSplit_res.csv')
    pos = ['pos']
    ts = ARCOS(df_in, posCols=pos, frame_column='t', id_column='id', measurement_column=None, clid_column='collid')
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['pos'])
    assert_frame_equal(out, df_true)


def test_split_from_2_objects():
    df_in = pd.read_csv('tests/testdata/2objSplit_in.csv')
    df_true = pd.read_csv('tests/testdata/2objSplit_res.csv')
    pos = ['pos']
    ts = ARCOS(df_in, posCols=pos, frame_column='t', id_column='id', measurement_column=None, clid_column='collid')
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['pos'])
    assert_frame_equal(out, df_true)


def test_cross_2_objects():
    df_in = pd.read_csv('tests/testdata/2objCross_in.csv')
    df_true = pd.read_csv('tests/testdata/2objCross_res.csv')
    pos = ['pos']
    ts = ARCOS(df_in, posCols=pos, frame_column='t', id_column='id', measurement_column=None, clid_column='collid')
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['pos'])
    assert_frame_equal(out, df_true)


def test_merge_split_2_objects_with_common():
    df_in = pd.read_csv('tests/testdata/2objMergeSplitCommon_in.csv')
    df_true = pd.read_csv('tests/testdata/2objMergeSplitCommon_res.csv')
    pos = ['pos']
    ts = ARCOS(df_in, posCols=pos, frame_column='t', id_column='id', measurement_column=None, clid_column='collid')
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['pos'])
    assert_frame_equal(out, df_true)


## if algorithm behaves differently (like in R) a different output is produced regarding collective events
def test_merge_split_2_objects_crossing():
    df_in = pd.read_csv('tests/testdata/2objMergeSplitCross_in.csv')
    df_true = pd.read_csv('tests/testdata/2objMergeSplitCross_res.csv')
    pos = ['pos']
    ts = ARCOS(df_in, posCols=pos, frame_column='t', id_column='id', measurement_column=None, clid_column='collid')
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['pos'])
    assert_frame_equal(out, df_true)


## if algorithm behaves differently (like in R) a different output is produced regarding collective events
def test_merge_and_split_2_objects_near():
    df_in = pd.read_csv('tests/testdata/2objMergeSplitNear_in.csv')
    df_true = pd.read_csv('tests/testdata/2objMergeSplitNear_res.csv')
    pos = ['pos']
    ts = ARCOS(df_in, posCols=pos, frame_column='t', id_column='id', measurement_column=None, clid_column='collid')
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['pos'])
    assert_frame_equal(out, df_true)


def test_4_objects_in_2_events():
    df_in = pd.read_csv('tests/testdata/4obj2events_in.csv')
    df_true = pd.read_csv('tests/testdata/4obj2events_res.csv')
    pos = ['x']
    ts = ARCOS(df_in, posCols=pos, frame_column='frame', id_column='id', measurement_column=None, clid_column='collId')
    out = ts.trackCollev(eps=1.0, minClsz=1, nPrev=1)
    out = out.drop(columns=['x'])
    assert_frame_equal(out, df_true)



"""Module containing binarization and detrending classes.

Example:
    >>> from arcos4py.tools import binData
    >>> binarizer = binData(biasMet="lm", polyDeg=1)
    >>> data_rescaled = binarizer.run(data, colMeas="ERK_KTR", colGroup="trackID")
"""
import numpy as np
import pandas as pd
from scipy.ndimage import median_filter
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures, minmax_scale


class detrender:
    """Smooth and de-trend input data.

    First, a short-term median filter with size smoothK is applied
    to remove fast noise from the time series.
    The subsequent de-trending can be performed with a long-term median filter
    with the size biasK {biasMet = "runmed"}
    or by fitting a polynomial of degree polyDeg {biasMet = "lm"}.

    Attributes:
        smoothK (int): Representing the size of the short-term median smoothing filter.
        biasK (int): Representing the size of the long-term de-trending median filter.
        peakThr (float): Threshold for rescaling of the de-trended signal.
        polyDeg (int): Sets the degree of the polynomial for lm fitting.
        biasMet (str): Indicating de-trending method, one of ['runmed', 'lm', 'none'].
    """

    def __init__(
        self,
        smoothK: int = 3,
        biasK: int = 51,
        peakThr: float = 0.2,
        polyDeg: int = 1,
        biasMet: str = "runmed",
    ) -> None:
        """Smooth and de-trend input data.

        Arguments:
            smoothK (int): Representing the size of the short-term median smoothing filter.
            biasK (int): Representing the size of the long-term de-trending median filter.
            peakThr (float): Threshold for rescaling of the de-trended signal.
            polyDeg (int): Sets the degree of the polynomial for lm fitting.
            biasMet (str): Indicating de-trending method, one of ['runmed', 'lm', 'none'].
        """
        # check if biasmethod contains one of these three types
        biasMet_types = ["runmed", "lm", "none"]
        if biasMet not in biasMet_types:
            raise ValueError(f"Invalid bias method. Expected one of: {biasMet_types}")

        self.smoothK = smoothK
        self.biasK = biasK
        self.peakThr = peakThr
        self.polyDeg = polyDeg
        self.biasMet = biasMet

    def _detrend_runnmed(self, x, filter_size, endrule_mode):
        local_smoothing = median_filter(input=x, size=filter_size, mode=endrule_mode)
        return local_smoothing

    # solution a bit janky but only median filter with changing window size?
    def _detrend_global_runmed(self, x, filter_size):
        x_series = pd.Series(x)
        global_smoothing = x_series.rolling(filter_size, min_periods=1).median().to_numpy()
        return global_smoothing

    def _detrend_lm(self, x, polynomial_degree):
        y = np.linspace(1, x.size, x.size).astype(int).reshape((-1, 1))
        transformer = PolynomialFeatures(degree=polynomial_degree, include_bias=False)
        data_ = transformer.fit_transform(y)
        model = LinearRegression().fit(X=data_, y=x)
        predicted_value = model.predict(data_)
        return predicted_value

    def _run_detrend(self, x: np.ndarray) -> np.ndarray:
        if x.size:
            local_smoothed = self._detrend_runnmed(x, self.smoothK, "nearest")
            if self.biasMet != "none":
                if self.biasMet == "runmed":
                    global_smoothed = self._detrend_global_runmed(
                        x=local_smoothed,
                        filter_size=self.biasK,
                    )
                elif self.biasMet == "lm":
                    global_smoothed = self._detrend_lm(local_smoothed, self.polyDeg)

                local_smoothed = np.subtract(local_smoothed, global_smoothed)
                local_smoothed = np.clip(local_smoothed, 0, None)
                if (local_smoothed.max() - local_smoothed.min()) > self.peakThr:
                    local_smoothed = np.divide(local_smoothed, local_smoothed.max())
                local_smoothed = np.nan_to_num(local_smoothed)
        else:
            local_smoothed = np.array([])

        return local_smoothed

    def detrend(self, x: np.ndarray, group_index: int, meas_index: int) -> np.ndarray:
        """Run detrinding on input data.

        The method applies detrending to each group defined in group_col and
        outputs it into the resc_column.

        Arguments:
            x (np.ndarray): Time series data for smoothing.
            group_index (int): Index of id column in x.
            meas_index (int): Index of measurement column in x.

        Returns (np.ndarray): Dataframe containing rescaled column.
        """
        #################################################
        meas_array = x[:, meas_index].astype('float64')
        try:
            group_array = x[:, group_index].astype('int64')
        except ValueError:
            try:
                group_array = x[:, group_index].astype('float64')
            except ValueError:
                group_array = x[:, group_index].astype('U6')

        grouped_array = np.split(meas_array, np.unique(group_array, axis=0, return_index=True)[1][1:])
        out = [self._run_detrend(x) for x in grouped_array]
        out_list = [item for sublist in out for item in sublist]
        return np.array(out_list)


class binData(detrender):
    """Smooth, de-trend, and binarise the input data.

    First a short-term median filter with size smoothK
    is applied to remove fast noise from the time series.
    If the de-trending method is set to "none",
    smoothing is applied on globally rescaled time series.
    The subsequent de-trending can be performed with a long-term median filter
    with the size biasK {biasMet = "runmed"}
    or by fitting a polynomial of degree polyDeg {biasMet = "lm"}.

    After de-trending,
    if the global difference between min/max is greater than the threshold
    the signal is rescaled to the (0,1) range.
    The final signal is binarised using the binThr threshold.

    Attributes:
        smoothK (int): Size of the short-term median smoothing filter.
        biasK (int): Size of the long-term de-trending median filter.
        peakThr (float): Threshold for rescaling of the de-trended signal.
        binThr (float): Threshold for binarizing the de-trended signal.
        polyDeg (int): Sets the degree of the polynomial for lm fitting.
        biasMet (str): De-trending method, one of ['runmed', 'lm', 'none'].
    """

    def __init__(
        self,
        smoothK: int = 3,
        biasK: int = 51,
        peakThr: float = 0.2,
        binThr: float = 0.1,
        polyDeg: int = 1,
        biasMet: str = "runmed",
    ) -> None:
        """Smooth, de-trend, and binarise the input data.

        Arguments:
            smoothK (int): Size of the short-term median smoothing filter.
            biasK (int): Size of the long-term de-trending median filter.
            peakThr (float): Threshold for rescaling of the de-trended signal.
            binThr (float): Threshold for binarizing the de-trended signal.
            polyDeg (int): Sets the degree of the polynomial for lm fitting.
            biasMet (str): De-trending method, one of ['runmed', 'lm', 'none'].
        """
        super().__init__(smoothK, biasK, peakThr, polyDeg, biasMet)
        self.binThr = binThr

    def _rescale_data(self, x: np.ndarray, group_index: int, meas_index: int, feat_range: tuple = (0, 1)) -> np.ndarray:
        meas_array = x[:, meas_index].astype('float64')
        try:
            group_array = x[:, group_index].astype('int64')
        except ValueError:
            try:
                group_array = x[:, group_index].astype('float64')
            except ValueError:
                group_array = x[:, group_index].astype('U6')

        grouped_array = np.split(meas_array, np.unique(group_array, axis=0, return_index=True)[1][1:])
        out = [minmax_scale(i, feature_range=feat_range) if i.shape[0] > 0 else np.array([]) for i in grouped_array]
        rescaled = [item for sublist in out for item in sublist]
        # rescaled = minmax_scale(x[:,1], feature_range=feat_range)
        x[:, meas_index] = rescaled
        return x

    def _bin_data(self, x: np.ndarray) -> np.ndarray:
        bin = (x >= self.binThr).astype(np.int_)
        return bin

    def run(self, x: pd.DataFrame, colGroup: str, colMeas: str, colFrame: str) -> pd.DataFrame:
        """Runs binarization and detrending.

        If the bias Method is 'none', first it rescales the data to between [0,1], then
        local smoothing is applied to the measurement by groups, followed by
        binarization.

        If biasMeth is one of ['lm', 'runmed'], first the data is detrended locally with a
        median filter and then detrended globally, for 'lm' with a linear model and for 'runmed' with a
        median filter.
        Followed by binarization of the data.

        Arguments:
            x (DataFrame): The time-series data for smoothing, detrending and binarization.
            colGroup (str): Object id column in x. Detrending and rescaling is performed on a per-object basis.
            colMeas (str): Measurement column in x on which detrending and rescaling is performed.
            colFrame (str): Frame column in Time-series data. Used for sorting.

        Returns:
            DataFrame: Dataframe containing binarized data, rescaled data and the original columns.
        """
        col_resc = f"{colMeas}.resc"
        col_bin = f"{colMeas}.bin"
        col_fact = f'{colGroup}_factorized'
        cols = [col_fact, colMeas]
        x.sort_values([colGroup, colFrame], inplace=True)
        # factorize column in order to prevent numpy grouping error in detrending
        value, label = x[colGroup].factorize()
        x[col_fact] = value
        # convert to numpy
        data_np = x[cols].to_numpy()

        if self.biasMet == "none":
            rescaled_data = self._rescale_data(data_np, group_index=0, meas_index=1)
            detrended_data = self.detrend(rescaled_data, 0, 1)
            binarized_data = self._bin_data(detrended_data)
        else:
            detrended_data = self.detrend(data_np, group_index=0, meas_index=1)
            binarized_data = self._bin_data(detrended_data)

        x = x.drop([col_fact], axis=1)
        x[col_resc] = detrended_data
        x[col_bin] = binarized_data
        return x



import pandas as pd
import pytest
from numpy import NaN, int64
from pandas.testing import assert_frame_equal

from arcos4py.tools import interpolation


@pytest.fixture
def test_data() -> pd.DataFrame:
    d = {
        't': [i if i != 4 else NaN for i in range(1, 8)],
        'y': [i for i in range(1, 8)],
        'id': [1 for i in range(7)],
    }
    df = pd.DataFrame(d)

    d2 = {
        't': [i for i in range(1, 8)],
        'y': [i for i in range(1, 8)],
        'id': [1 for i in range(7)],
    }
    df2 = pd.DataFrame(d2)
    return df, df2


def test_interpolate_middle(test_data: pd.DataFrame):
    df, df_out = test_data
    interpolated_t = interpolation(df['t']).interpolate()
    df['t'] = interpolated_t
    df['t'] = df['t'].astype(int64)
    assert_frame_equal(df, df_out)



"""Main Module of arcos4py.

This module contains the ARCOS class, which implements most functionallity of arcos4py
to prepare data and to detect and track collective events.

Example:
    >>> from arcos4py import ARCOS
    >>> ts = ARCOS(data,["x"], 'time', 'id', 'meas', 'clTrackID')
    >>> ts.interpolate_measurements()
    >>> ts.clip_meas(clip_low = 0.001, clip_high=0.999)
    >>> ts.bin_measurements(
            smoothK = 3,
            biasK = 51,
            peakThr = 0.2,
            binThr = 0.1,
            polyDeg = 1,
            biasMet = "runmed")
    >>> events_df = ts.trackCollev(eps = 1, minClsz = 1, nPrev = 1)
"""

from typing import Union

import pandas as pd

from .tools.binarize_detrend import binData
from .tools.cleandata import clipMeas, interpolation
from .tools.detect_events import detectCollev


class ARCOS:
    """Detects and tracks collective events in a tracked time-series dataset.

    Requires binarized measurement column, that can be generated with the
    bin_measurements method.
    Tracking makes use of the dbscan algorithm, which is applied to every frame
    and subsequently connects collective events between frames located
    within eps distance of each other.

    Attributes:
        data (DataFrame): Data of tracked time-series in "long format". Can be used to
            acess modified dataframe at any point.
        posCols (list): List containing position column names strings inside data e.g.
            At least one dimension is required.
        frame_column (str): Indicating the frame column in input_data.
        id_column (str): Indicating the track id/id column in input_data.
        measurement_column (str): Indicating the measurement column in input_data.
        clid_column (str): Indicating the column name containing the collective event ids.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        posCols: list = ["x"],
        frame_column: str = 'time',
        id_column: str = 'id',
        measurement_column: str = 'meas',
        clid_column: str = 'clTrackID',
    ) -> None:
        """Constructs class with provided arguments.

        Arguments:
            data (DataFrame): Input Data of tracked time-series in "long format" containing position columns,
                a measurement and an object ID column.
            posCols (list): List ontaining position column names strings inside data e.g.
                At least one dimension is required.
            frame_column (str): Indicating the frame column in input_data.
            id_column (str): Indicating the track id/object id column in input_data.
            measurement_column (str): Indicating the measurement column in input_data.
            clid_column (str): Indicating the column name containing the collective event ids.
        """
        self.data = data
        self.posCols = posCols
        self.frame_column = frame_column
        self.id_column = id_column
        self.measurement_column = measurement_column
        self.clid_column = clid_column

        self.data_binarized: pd.DataFrame = None
        self.tracked_events: pd.DataFrame = None
        self.bin_col: Union[str, None] = None
        # to check if no measurement was provided assign None

        self.data = self.data.sort_values(by=[self.frame_column, self.id_column])
        self._check_col()
        if self.measurement_column is not None:
            self.resc_col = f"{self.measurement_column}.resc"
            self.bin_col = f"{self.measurement_column}.bin"

    def __repr__(self) -> pd.DataFrame:
        """Set __repr___ to return self.data."""
        return repr(self.data)

    def _check_col(self):
        """Checks that self.cols contains all required columns."""
        columns = self.data.columns
        input_columns = [self.frame_column, self.id_column, self.id_column, self.measurement_column]
        input_columns = [col for col in input_columns if col is not None]
        if not all(item in columns for item in input_columns):
            raise ValueError(f"Columns {input_columns} do not match with column in dataframe.")

    def interpolate_measurements(self) -> pd.DataFrame:
        """Interpolates NaN's in place in measurement column.

        Returns:
            Dataframe with interpolated measurement column.
        """
        meas_interp = interpolation(self.data).interpolate()
        self.data = meas_interp
        return self.data

    def clip_meas(self, clip_low: float = 0.001, clip_high: float = 0.999) -> pd.DataFrame:
        """Clip measurement column to upper and lower quantiles defined in clip_low and clip_high.

        Arguments:
            clip_low (float): Lower clipping boundary (quantile).

            clip_high (float): Upper clipping boundary (quantile).

        Returns:
            Dataframe with in place clipped measurement column.
        """
        meas_column = self.data[self.measurement_column].to_numpy()
        meas_clipped = clipMeas(meas_column).clip(clip_low, clip_high)
        self.data[self.measurement_column] = meas_clipped
        return self.data

    def bin_measurements(
        self,
        smoothK: int = 3,
        biasK: int = 51,
        peakThr: float = 0.2,
        binThr: float = 0.1,
        polyDeg: int = 1,
        biasMet: str = "runmed",
    ) -> pd.DataFrame:
        r"""Smooth, de-trend, and binarise the input data.

        First a short-term median filter with size smoothK
        is applied to remove fast noise from the time series.
        If the de-trending method is set to "none",
        smoothing is applied on globally rescaled time series.
        The subsequent de-trending can be performed with a long-term median filter
        with the size biasK {biasMet = "runmed"}
        or by fitting a polynomial of degree polyDeg {biasMet = "lm"}.

        After de-trending,
        if the global difference between min/max is greater than the threshold
        the signal is rescaled to the (0,1) range.
        The final signal is binarised using the binThr threshold

        Arguments:
            smoothK (int): Size of the short-term median smoothing filter.
            biasK (int): Size of the long-term de-trending median filter
            peakThr (float): Threshold for rescaling of the de-trended signal.
            binThr (float): Threshold for binary classification.
            polyDeg (int): Sets the degree of the polynomial for lm fitting.
            biasMet (str): De-trending method, one of ['runmed', 'lm', 'none'].

        Returns:
            DataFrame with detrended/smoothed and binarized measurement column.
        """
        self.data = binData(
            smoothK,
            biasK,
            peakThr,
            binThr,
            polyDeg,
            biasMet,
        ).run(self.data, colMeas=self.measurement_column, colGroup=self.id_column, colFrame=self.frame_column)
        return self.data

    def trackCollev(self, eps: float = 1, minClsz: int = 1, nPrev: int = 1) -> pd.DataFrame:
        """Requires binarized measurement column.

        Makes use of the dbscan algorithm,
        applies this to every timeframe and subsequently connects
        collective events between frames located within eps distance of each other.

        Arguments:
            eps (float): The maximum distance between two samples for one to be considered as in
                the neighbourhood of the other.
                This is not a maximum bound on the distances of points within a cluster.
                Value is also used to connect collective events across multiple frames.
            minClsz (str): The minimum size for a cluster to be identified as a collective event
            nPrev (int): Number of previous frames the tracking
                algorithm looks back to connect collective events

        Returns:
            DataFrame with detected collective events across time.
        """
        self.data = detectCollev(
            self.data,
            eps=eps,
            minClSz=minClsz,
            nPrev=nPrev,
            posCols=self.posCols,
            frame_column=self.frame_column,
            id_column=self.id_column,
            bin_meas_column=self.bin_col,
            clid_column=self.clid_column,
        ).run()

        return self.data



"""Module to track and detect collective events.

Example:
    >>> from arcos4py.tools import detectCollev
    >>> ts = detectCollev(data)
    >>> events_df = ts.run()
"""

from typing import Union

import numpy as np
import pandas as pd
from scipy.spatial import KDTree
from sklearn.cluster import DBSCAN

from ._errors import columnError, epsError, minClSzError, noDataError, nPrevError


class detectCollev:
    """Identifies and tracks collective signalling events.

    Requires binarized measurement column.
    Makes use of the dbscan algorithm,
    applies this to every timeframe and subsequently connects
    collective events between frames located within eps distance of each other.

    Attributes:
        input_data (DataFrame): Input data to be processed. Must contain a binarized measurement column.
        eps (float): The maximum distance between two samples for one to be considered as in
            the neighbourhood of the other.
            This is not a maximum bound on the distances of points within a cluster.
            Value is also used to connect collective events across multiple frames.
        minClSz (int): Minimum size for a cluster to be identified as a collective event.
        nPrev (int): Number of previous frames the tracking
            algorithm looks back to connect collective events.
        posCols (list): List of position columns contained in the data.
            Must at least contain one
        frame_column (str): Indicating the frame column in input_data.
        id_column (str): Indicating the track id/id column in input_data.
        bin_meas_column (str): Indicating the bin_meas_column in input_data or None.
        clid_column (str): Indicating the column name containing the ids of collective events.
    """

    def __init__(
        self,
        input_data: pd.DataFrame,
        eps: float = 1,
        minClSz: int = 1,
        nPrev: int = 1,
        posCols: list = ["x"],
        frame_column: str = 'time',
        id_column: Union[str, None] = None,
        bin_meas_column: Union[str, None] = 'meas',
        clid_column: str = 'clTrackID',
    ) -> None:
        """Constructs class with input parameters.

        Arguments:
            input_data (DataFrame): Input data to be processed. Must contain a binarized measurement column.
            eps (float): The maximum distance between two samples for one to be considered as in
                the neighbourhood of the other.
                This is not a maximum bound on the distances of points within a cluster.
                Value is also used to connect collective events across multiple frames.
            minClSz (int): Minimum size for a cluster to be identified as a collective event.
            nPrev (int): Number of previous frames the tracking
                algorithm looks back to connect collective events.
            posCols (list): List of position columns contained in the data.
                Must at least contain one
            frame_column (str): Indicating the frame column in input_data.
            id_column (str | None): Indicating the track id/id column in input_data, optional.
            bin_meas_column (str): Indicating the bin_meas_column in input_data or None.
            clid_column (str): Indicating the column name containing the ids of collective events.
        """
        # assign some variables passed in as arguments to the object
        self.input_data = input_data
        self.eps = eps
        self.minClSz = minClSz
        self.nPrev = nPrev
        self.frame_column = frame_column
        self.id_column = id_column
        self.bin_meas_column = bin_meas_column
        self.clid_column = clid_column
        self.posCols = posCols
        self.columns_input = self.input_data.columns
        self.clidFrame = f'{clid_column}.frame'

        self.pos_cols_inputdata = [col for col in self.posCols if col in self.columns_input]

        # run input checks
        self._run_input_checks()

    def _check_input_data(self):
        """Checks if input contains data\
        raises error if not."""
        if self.input_data is None:
            raise noDataError("Input is None")
        elif self.input_data.empty:
            raise noDataError("Input is empty")

    def _check_pos_columns(self):
        """Checks if Input contains correct columns\
        raises Exception if not."""
        if not all(item in self.columns_input for item in self.posCols):
            raise columnError("Input data does not have the indicated position columns!")

    def _check_frame_column(self):
        if self.frame_column not in self.columns_input:
            raise columnError("Input data does not have the indicated frame column!")

    def _check_eps(self):
        """Checks if eps is greater than 0."""
        if self.eps <= 0:
            raise epsError("eps has to be greater than 0")

    def _check_minClSz(self):
        """Checks if minClSiz is greater than 0."""
        if self.minClSz <= 0:
            raise minClSzError("Parameter minClSiz has to be greater than 0!")

    def _check_nPrev(self):
        """Checks if nPrev is greater than 0."""
        if self.nPrev <= 0 and isinstance(self.nPrev, int):
            raise nPrevError("Parameter nPrev has to be an integer greater than 0 and an integer!")

    def _run_input_checks(self):
        """Run input checks."""
        self._check_input_data()
        self._check_pos_columns()
        self._check_eps()
        self._check_minClSz()
        self._check_nPrev()
        self._check_frame_column()

    def _select_necessary_columns(
        self, data: pd.DataFrame, frame_col: str, id_col: Union[str, None], pos_col: list, bin_col: Union[str, None]
    ) -> pd.DataFrame:
        """Select necessary input colums from input data into dataframe.

        Arguments:
            data (DataFrame): Containing necessary columns.
            frame_col (str): Frame column in data.
            id_col (str): Id column in data.
            pos_col (list): string representation of position columns in data.
            bin_col (str): Name of binary column.

        Returns:
            DataFrame: Filtered columns necessary for calculation.
        """
        columns = [frame_col, id_col, bin_col]
        columns = [col for col in columns if col]
        columns.extend(pos_col)
        neccessary_data = data[columns].copy(deep=True)
        return neccessary_data

    def _filter_active(self, data: pd.DataFrame, bin_meas_col: Union[str, None]) -> pd.DataFrame:
        """Selects rows with binary value of greater than 0.

        Arguments:
            data (DataFrame): Dataframe containing necessary columns.
            bin_meas_col (str|None): Either name of the binary column or None if no such column exists.

        Returns:
            DataFrame: Filtered pandas DataFrame.
        """
        if bin_meas_col is not None:
            data = data[data[bin_meas_col] > 0]
        return data

    def _dbscan(self, x: np.ndarray) -> list:
        """Dbscan method to run and merge the cluster id labels to the original dataframe.

        Arguments:
            x (np.ndarray): With unique frame and position columns.
            collid_col (str): Column to be created containing cluster-id labels.

        Returns:
            list[np.ndarray]: list with added collective id column detected by DBSCAN.
        """
        if x.size:
            db_array = DBSCAN(eps=self.eps, min_samples=self.minClSz, algorithm="kd_tree").fit(x[:, 1:])
            cluster_labels = db_array.labels_
            cluster_list = [id + 1 if id > -1 else np.nan for id in cluster_labels]
            return cluster_list

        return []

    def _run_dbscan(self, data: pd.DataFrame, frame: str, clid_frame: str, id_column: Union[str, None]) -> pd.DataFrame:
        """Apply dbscan method to every group i.e. frame.

        Arguments:
            data (DataFrame): Must contain position columns and frame columns.
            frame (str): Name of frame column in data.
            clid_frame (str): column to be created containing the output cluster ids from dbscan.
            id_column (str | None): track_id column

        Returns:
            DataFrame: Dataframe with added collective id column detected by DBSCAN for every frame.
        """
        if self.id_column:
            data = data.sort_values([frame, id_column]).reset_index(drop=True)
        else:
            data = data.sort_values([frame]).reset_index(drop=True)
        subset = [frame] + self.pos_cols_inputdata
        data_np = data[subset].to_numpy(dtype=np.float64)
        grouped_array = np.split(data_np, np.unique(data_np[:, 0], axis=0, return_index=True)[1][1:])
        # map dbscan to grouped_array
        out = [self._dbscan(i) for i in grouped_array]
        out_list = [item for sublist in out for item in sublist]
        data[clid_frame] = out_list
        data = data.dropna()
        return data

    def _make_db_id_unique(self, db_data: pd.DataFrame, frame: str, clid_frame, clid) -> pd.DataFrame:
        """Make db_scan cluster id labels unique by adding the\
        cummulative sum of previous group to next group.

        Arguments:
            db_data (DataFrame): Returned by _run_dbscan function with non-unique cluster ids.
            frame (str): Frame column.
            clid_frame (str): Column name of cluster-id per frame.
            clid (str): Column name of unique cluster ids to be returned.

        Returns:
            DataFrame: Dataframe with unique collective events.
        """
        db_data_np = db_data[[frame, clid_frame]].to_numpy()
        grouped_array = np.split(db_data_np[:, 1], np.unique(db_data_np[:, 0], axis=0, return_index=True)[1][1:])
        max_array = [0] + [np.max(i) for i in grouped_array if i.size != 0]
        out = [np.add(value, np.cumsum(max_array)[i]) for i, value in enumerate(grouped_array)]
        db_gp = np.concatenate(out)
        db_data[clid] = db_gp.astype(np.int64)
        return db_data

    def _nearest_neighbour(
        self,
        data_a: np.ndarray,
        data_b: np.ndarray,
        nbr_nearest_neighbours: int = 1,
    ):
        """Calculates nearest neighbour in from data_a\
        to data_b nearest_neighbours in data_b.

        Arguments:
            data_a (DataFrame): containing position values.
            data_b (DataFrame): containing position values.
            nbr_nearest_neighbours (int): of the number of nearest neighbours to be calculated.

        Returns:
            tuple(np.ndarray, np.ndarray): Returns tuple of 2 arrays containing nearest neighbour indices and distances.
        """
        kdB = KDTree(data=data_a)
        nearest_neighbours = kdB.query(data_b, k=nbr_nearest_neighbours)
        return nearest_neighbours

    def _link_clusters_between_frames(self, data: pd.DataFrame, frame: str, colid: str) -> pd.DataFrame:
        """Tracks clusters detected with DBSCAN along a frame axis,\
        returns tracked collective events as a pandas dataframe.

        Arguments:
            data (DataFrame): Output from dbscan.
            frame (str): Frame column.
            colid (str): Colid column.

        Returns:
            DataFrame: Pandas dataframe with tracked collective ids.
        """
        essential_cols = [frame, colid] + self.posCols
        data_essential = data[essential_cols]
        data_np = data_essential.to_numpy()
        data_np_frame = data_np[:, 0]

        # loop over all frames to link detected clusters iteratively
        for t in np.unique(data_np_frame, return_index=False)[1:]:
            prev_frame = data_np[(data_np_frame >= (t - self.nPrev)) & (data_np_frame < t)]
            current_frame = data_np[data_np_frame == t]
            # only continue if objects were detected in previous frame
            if prev_frame.size:
                colid_current = current_frame[:, 1]
                # loop over unique cluster in frame
                for cluster in np.unique(colid_current, return_index=False):
                    pos_current = current_frame[:, 2:][colid_current == cluster]
                    pos_previous = prev_frame[:, 2:]
                    # calculate nearest neighbour between previoius and current frame
                    nn_dist, nn_indices = self._nearest_neighbour(pos_previous, pos_current)
                    prev_cluster_nbr_all = prev_frame[nn_indices, 1]
                    prev_cluster_nbr_eps = prev_cluster_nbr_all[(nn_dist <= self.eps)]
                    # only continue if neighbours
                    # were detected within eps distance
                    if prev_cluster_nbr_eps.size:
                        prev_clusternbr_eps_unique = np.unique(prev_cluster_nbr_eps, return_index=False)
                        if prev_clusternbr_eps_unique.size > 0:
                            # propagate cluster id from previous frame
                            data_np[((data_np_frame == t) & (data_np[:, 1] == cluster)), 1] = prev_cluster_nbr_all

        np_out = data_np[:, 1]
        sorter = np_out.argsort()[::1]
        grouped_array = np.split(np_out[sorter], np.unique(np_out[sorter], axis=0, return_index=True)[1][1:])
        np_grouped_consecutive = (np.repeat(i + 1, value.size) for i, value in enumerate(grouped_array))
        out_array = np.array([item for sublist in np_grouped_consecutive for item in sublist])
        data[colid] = out_array[sorter.argsort()].astype('int64')
        return data

    def _get_export_columns(self):
        """Get columns that will contained in the pandas dataframe returned by the run method."""
        self.pos_cols_inputdata = [col for col in self.posCols if col in self.columns_input]
        if self.id_column:
            columns = [self.frame_column, self.id_column]
        else:
            columns = [self.frame_column]
        columns.extend(self.pos_cols_inputdata)
        columns.append(self.clid_column)
        return columns

    def run(self) -> pd.DataFrame:
        """Method to execute the different steps necessary for tracking.

        1. Selects columns.
        2. filters data on binary column > 1.
        3. Applies dbscan algorithm to every frame.
        4. Makes cluster ids unique across frames.
        5. Tracks collective events i.e. links cluster ids across frames.
        6. Creates final DataFrame.

        Returns:
            DataFrame: Dataframe with tracked collective events is returned.
        """
        filtered_cols = self._select_necessary_columns(
            self.input_data,
            self.frame_column,
            self.id_column,
            self.pos_cols_inputdata,
            self.bin_meas_column,
        )
        active_data = self._filter_active(filtered_cols, self.bin_meas_column)
        db_data = self._run_dbscan(
            data=active_data,
            frame=self.frame_column,
            clid_frame=self.clidFrame,
            id_column=self.id_column,
        )
        db_data = self._make_db_id_unique(
            db_data,
            frame=self.frame_column,
            clid_frame=self.clidFrame,
            clid=self.clid_column,
        )
        tracked_events = self._link_clusters_between_frames(db_data, self.frame_column, self.clid_column)
        return_columns = self._get_export_columns()
        tracked_events = tracked_events[return_columns]
        if self.clid_column in self.input_data.columns:
            df_to_merge = self.input_data.drop(columns=[self.clid_column])
        else:
            df_to_merge = self.input_data
        tracked_events = tracked_events.merge(df_to_merge, how="left")
        tracked_events = tracked_events
        return tracked_events



"""Module containing clipping and interpolation classes.

Example:
    >>> # Interpolation
    >>> from arcos4py.tools import interpolation
    >>> a = interpolation(data)
    >>> data_interp = a.interpolate()

    >>> # clipping
    >>> from arcos4py.tools import clipMeas
    >>> a = clipMeas(data)
    >>> data_clipped = a.clip(0.001, 0.999)

"""
import numpy as np
import pandas as pd


class interpolation:
    """Interpolate nan values in a numpy array.

    Attributes:
        data (DataFrame): Where NaN should be replaced with interpolated values.
    """

    def __init__(self, data: pd.DataFrame):
        """Interpolate nan values in a pandas dataframe.

        Uses pandas.interpolate with liner interpolation.

        Arguments:
            data (DataFrame): Where NaN should be replaced with interpolated values.
        """
        self.data = data

    def interpolate(self) -> pd.DataFrame:
        """Interpolate nan and missing values.

        Returns:
            DataFrame: Interpolated input data.
        """
        self.data = self.data.interpolate(axis=0)

        return self.data


class clipMeas:
    """Clip input array."""

    def __init__(self, data: np.ndarray) -> None:
        """Clips array to quantilles.

        Arguments:
            data (ndarray): To be clipped.
        """
        self.data = data

    def _calculate_percentile(self, data: np.ndarray, clip_low: float, clip_high: float):
        """Calculate upper and lower quantille.

        Arguments:
            data (ndarray): To calculate upper and lower quantile on.
            clip_low (float): Lower clipping boundary (quantile).
            clip_high (float): Upper clipping boundry (quantille).

        Returns:
            np.ndarray: Array with lower quantile and array with upper quantile.

        """
        quantille_low = np.quantile(data, clip_low, keepdims=True)
        quantille_high = np.quantile(data, clip_high, keepdims=True)
        return quantille_low, quantille_high

    def clip(self, clip_low: float = 0.001, clip_high: float = 0.999) -> np.ndarray:
        """Clip input array to upper and lower quantiles defined in clip_low and clip_high.

        Arguments:
            clip_low (float): Lower clipping boundary (quantile).
            clip_high (float): Upper clipping boundry (quantille).

        Returns:
            np.ndarray: A clipped array of the input data.
        """
        low, high = self._calculate_percentile(self.data, clip_low, clip_high)
        out = self.data.clip(low, high)
        return out



class noDataError(ValueError):
    """Raised when input Data is None or empty."""


class columnError(ValueError):
    """Raised when columns passed to track_collev\
    are not present in data."""


class epsError(ValueError):
    """Raised if eps is smaller than 1."""


class minClSzError(ValueError):
    """Raised if minClSz is smaller than 1."""


class nPrevError(ValueError):
    """Raised if nPrev is smaller than 1."""



"""Tools for detecting collective events."""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = '0.1.5'

from .binarize_detrend import binData
from .cleandata import clipMeas, interpolation
from .detect_events import detectCollev
from .filter_events import filterCollev
from .stats import calcCollevStats

__all__ = [
    "binData",
    "clipMeas",
    "interpolater",
    "detectCollev",
    "filterCollev",
    "calcCollevStats",
]



"""Module to filter collective events.

Example:
    >>> from arcos4py.tools import filterCollev
    >>> f = filterCollev(data, 'time', 'collid')
    >>> df = f.filter(coll_duration = 9, coll_total_size = 10)
"""
import pandas as pd

from .stats import calcCollevStats


class filterCollev:
    """Select Collective events that last longer than coll_duration\
    and have a larger total size than coll_total_size.

    Attributes:
        data (Dataframe): With detected collective events.
        frame_column (str): Indicating the frame column in data.
        collid_column (str): Indicating the collective event id column in data.
        obj_id_column (str): Inidicating the object identifier column such as cell track id.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        frame_column: str = "time",
        collid_column: str = "collid",
        obj_id_column: str = "trackID",
    ):
        """Constructs filterCollev class with Parameters.

        Arguments:
            data (Dataframe): With detected collective events.
            frame_column (str): Indicating the frame column in data.
            collid_column (str): Indicating the collective event id column in data.
            obj_id_column (str): Inidicating the object identifier column such as cell track id.
        """
        self.data = data
        self.frame_column = frame_column
        self.collid_column = collid_column
        self.obj_id_column = obj_id_column

    def _filter_collev(
        self,
        data: pd.DataFrame,
        collev_stats: pd.DataFrame,
        collev_id: str,
        min_duration: int,
        min_size: int,
    ):
        """Uses the dataframe generated by self._get_collev_duration()\
        to filter collective events that last longer than\
        min_duration and are larger than min_size.

        Arguments:
            data (DataFrame): Containing unfiltered collective events.
            collev_stats (DataFrame): Containing stats of collective events.
            collev_id (str): Indicating the contained collective id column.
            min_duration (str): minimal duration of a collective event for it to be returned.
            min_size (int): minimal size for a collective event to be returned.

        Returns:
            DataFrame: Dataframe containing filtered collective events.

        """
        collev_stats = collev_stats[
            (collev_stats["duration"] >= min_duration) & (collev_stats["total_size"] >= min_size)
        ]
        data = data[data[collev_id].isin(collev_stats[collev_id])]
        return data

    def filter(self, coll_duration: int = 9, coll_total_size: int = 10) -> pd.DataFrame:
        """Filter collective events.

        Method to filter collective events according to the
        parameters specified in the object instance.

        Arguments:
            coll_duration (int): Minimal duration of collective events to be selected.
            coll_total_size (int): Minimal total size of collective events to be selected.

        Returns:
             Returns pandas dataframe containing filtered collective events
        """
        if self.data.empty:
            return self.data
        stats = calcCollevStats()
        colev_duration = stats.calculate(self.data, self.frame_column, self.collid_column, self.obj_id_column)

        filtered_df = self._filter_collev(
            self.data,
            colev_duration,
            self.collid_column,
            coll_duration,
            coll_total_size,
        )
        return filtered_df



"""Arcos4py top level module.

This package is a python port of an R package
to detect and analyse collective phenomena.
"""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = '0.1.5'

from .arcos4py import ARCOS

__all__ = ["ARCOS"]



"""Module to plot different metrics generated by arcos4py functions.

Examples:
    >>> # Data Plots
    >>> from arcos4py.plotting import dataPlots
    >>> data_plots = dataPlots(df,'time', 'meas', 'track_id')
    >>> hist = data_plots.histogram()
    >>> dens = data_plots.density_plot()
    >>> xt_plot = data_plots.position_t_plot({'x'}, n=20)

    >>> # Detrended vs original plot
    >>> from arcos4py.plotting import plotOriginalDetrended
    >>> plot = arcosPlots(data, 'time', 'meas', 'detrended', 'id')
    >>> plot.plot_detrended()

    >>> # Stats Plot
    >>> from arcos4py.plotting import statsPlots
    >>> coll_dur_size_scatter = statsPlots(stats).plot_events_duration('total_size','duration')

    >>> # Noodle Plot
    >>> from arcos4py.plotting import NoodlePlot
    >>> ndl = NoodlePlot(df,"collid", 'track_id', 'time', 'x', 'y')
    >>> ndl_plot = ndl.plot('x')
"""

from __future__ import annotations

from typing import Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

TAB20 = [
    "#1f77b4",
    "#aec7e8",
    "#ff7f0e",
    "#ffbb78",
    "#2ca02c",
    "#98df8a",
    "#d62728",
    "#ff9896",
    "#9467bd",
    "#c5b0d5",
    "#8c564b",
    "#c49c94",
    "#e377c2",
    "#f7b6d2",
    "#7f7f7f",
    "#c7c7c7",
    "#bcbd22",
    "#dbdb8d",
    "#17becf",
    "#9edae5",
]


class dataPlots:
    """Plot different metrics of input data.

    Attributes:
        data (Dataframe): containing ARCOS data.
        frame (str): name of frame column in data.
        measurement (str): name of measurement column in data.
        id (str): name of track id column.
    """

    def __init__(self, data: pd.DataFrame, frame: str, measurement: str, id: str):
        """Plot different metrics such as histogram, position-t and density.

        Arguments:
            data (Dataframe): containing ARCOS data.
            frame (str): name of frame column in data.
            measurement (str): name of measurement column in data.
            id (str): name of track id column.
        """
        self.data = data
        self.id = id
        self.frame = frame
        self.measurement = measurement

    def position_t_plot(self, posCol: set[str] = {'x'}, n: int = 20):
        """Plots X and Y over T to visualize tracklength.

        Arguments:
            posCol (set): containing names of position columns in data.
            n (int): number of samples to plot.

        Returns:
            fig, axes: Matplotlib fig and axes of density plot.
        """
        sample = pd.Series(self.data[self.id].unique()).sample(n)
        pd_from_r_df = self.data.loc[self.data[self.id].isin(sample)]
        fig, axes = plt.subplots(1, len(posCol), figsize=(6, 3))
        for label, df in pd_from_r_df.groupby(self.id):
            for index, value in enumerate(posCol):
                if len(posCol) > 1:
                    df.plot(x=self.frame, y=value, ax=axes[index], legend=None)
                else:
                    df.plot(x=self.frame, y=value, ax=axes, legend=None)
        if len(posCol) > 1:
            for index, value in enumerate(posCol):
                axes[index].set_title(value)
        else:
            axes.set_title(value)
        return fig, axes

    def density_plot(self, *args, **kwargs):
        """Density plot of measurement.

        Uses Seaborn distplot to plot measurement density.

        Arguments:
            measurement_col (str): name of measurement column.
            *args (Any): arguments passed on to seaborn histplot function.
            **kwargs (Any): keyword arguments passed on to seaborn histplot function.

        Returns:
            FacetGrid: Seaborn FacetGrid of density density plot.
        """
        plot = sns.displot(
            self.data[self.measurement], kind="kde", palette="pastel", label=self.measurement, *args, **kwargs
        )
        # Plot formatting
        plt.legend(prop={'size': 10})
        plt.title('Density Plot of Measurement')
        plt.xlabel('Measurement')
        plt.ylabel('Density')
        return plot

    def histogram(self, bins: str = 'auto', *args, **kwargs):
        """Histogram of tracklenght.

        Uses seaborn histplot function to plot tracklenght histogram.

        Arguments:
            bins (str): number or width of bins in histogram
            *args (Any): arguments passed on to seaborn histplot function.
            **kwargs (Any): keyword arguments passed on to seaborn histplot function.

        Returns:
            AxesSubplot: Matplotlib AxesSubplot of histogram.
        """
        # Draw histogram
        track_length = self.data.groupby(self.id).size()
        axes = sns.histplot(track_length, label="Track Length", bins=bins, *args, **kwargs)
        # Plot formatting
        plt.title('Track length Histogram')
        axes.set_xlabel('Track Length')
        axes.set_ylabel('Count')
        return axes


class plotOriginalDetrended:
    """Plot different detrended vs original data.

    Attributes:
        data (Dataframe): containing ARCOS data.
        frame (str): name of frame column in data.
        measurement (str): name of measurement column in data.
        detrended (str): name of detrended column with detrended data.
        id (str): name of track id column.
    """

    def __init__(self, data: pd.DataFrame, frame: str, measurement: str, detrended: str, id: str):
        """Plot detrended vs original data.

        Arguments:
            data (Dataframe): containing ARCOS data.
            frame (str): name of frame column in data.
            measurement (str): name of measurement column in data.
            detrended (str): name of detrended column with detrended data.
            id (str): name of track id column.
        """
        self.data = data
        self.measurement = measurement
        self.detrended = detrended
        self.id = id
        self.frame = frame

    def plot_detrended(
        self, n_samples: int = 25, subplots: tuple = (5, 5), plotsize: tuple = (20, 10)
    ) -> matplotlib.axes.Axes:
        """Method to plot detrended vs original data.

        Arguments:
            n_samples (int): Number of tracks to plot.
            subplots (tuple): Number of subplots, should be approx. one per sample.
            plotsize (tuple): Size of generated plot.

        Returns:
            Fig, Axes: Matplotlib figure and axes2d of detrended vs original data.
        """
        vals = np.random.choice(self.data[self.id].unique(), n_samples, replace=False)
        self.data = self.data.set_index(self.id).loc[vals].reset_index()
        grouped = self.data.groupby(self.id)

        ncols = subplots[0]
        nrows = subplots[1]

        fig, axes2d = plt.subplots(nrows=nrows, ncols=ncols, figsize=plotsize, sharey=True)

        for (key, ax) in zip(grouped.groups.keys(), axes2d.flatten()):
            grouped.get_group(key).plot(x=self.frame, y=[self.measurement, self.detrended], ax=ax)
            ax.get_legend().remove()

        handles, labels = ax.get_legend_handles_labels()
        fig.legend(handles, labels, loc="lower right")

        return fig, axes2d


class statsPlots:
    """Plot data generated by the stats module.

    Attributes:
        data (DataFrame): containing ARCOS stats data.
    """

    def __init__(self, data: pd.DataFrame):
        """Plot detrended vs original data.

        Arguments:
            data (DataFrame): containing ARCOS stats data.
        """
        self.data = data

    def plot_events_duration(self, total_size: str, duration: str, point_size: int = 40, *args, **kwargs):
        """Scatterplot of collective event duration.

        Arguments:
            total_size (str): name of total size column.
            duration (str):, name of column with collective event duration.
            point_size (int): scatterplot point size.
            *args (Any): Arguments passed on to seaborn scatterplot function.
            **kwargs (Any): Keyword arguments passed on to seaborn scatterplot function.

        Returns:
            Axes: Matplotlib Axes object of scatterplot
        """
        plot = sns.scatterplot(x=self.data[total_size], y=self.data[duration], s=point_size, *args, **kwargs)
        return plot


class NoodlePlot:
    """Create Noodle Plot of cell tracks, colored by collective event id.

    Attributes:
        df (pd.DataFrame): DataFrame containing collective events from arcos.
        colev (str): Name of the collective event column in df.
        trackid (str): Name of the track column in df.
        frame (str): Name of the frame column in df.
        posx (str): Name of the X coordinate column in df.
        posy (str): Name of the Y coordinate column in df.
        posz (str): Name of the Z coordinate column in df,
            or None if no z column.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        colev: str,
        trackid: str,
        frame: str,
        posx: str,
        posy: str,
        posz: Union[str, None] = None,
    ):
        """Constructs class with given parameters.

        Arguments:
            df (pd.DataFrame): DataFrame containing collective events from arcos.
            colev (str): Name of the collective event column in df.
            trackid (str): Name of the track column in df.
            frame (str): Name of the frame column in df.
            posx (str): Name of the X coordinate column in df.
            posy (str): Name of the Y coordinate column in df.
            posz (str | None): Name of the Z coordinate column in df,
                or None if no z column.
        """
        self.df = df
        self.colev = colev
        self.trackid = trackid
        self.frame = frame
        self.posx = posx
        self.posy = posy
        self.posz = posz

    def _prepare_data_noodleplot(
        self,
        df: pd.DataFrame,
        color_cylce: list[str],
        colev: str,
        trackid: str,
        frame: str,
        posx: str,
        posy: str,
        posz: Union[str, None] = None,
    ):
        """From arcos collective event data,\
        generates a list of numpy arrays, one for each event.

        Arguments:
            df (pd.DataFrame): DataFrame containing collective events from arcos.
            color_cylce (list[str]): list of colors used to color trackid's
                for individual collective events.
            colev (str): Name of the collective event column in df.
            trackid (str): Name of the track column in df.
            frame: (str): Name of the frame column in df.
            posx (str): Name of the X coordinate column in df.
            posy (str): Name of the Y coordinate column in df.
            posz (str): Name of the Z coordinate column in df,
                or None if no z column.

        Returns:
            list[np.ndarray], np.ndarray: List of collective events data,
            and colors for each collective event.
        """
        # values need to be sorted to group with numpy
        df = df.sort_values([colev, trackid])
        if posz:
            array = df[[colev, trackid, frame, posx, posy, posz]].to_numpy()
        else:
            array = df[[colev, trackid, frame, posx, posy]].to_numpy()
        # generate goroups for each unique value
        grouped_array = np.split(array, np.unique(array[:, 0], axis=0, return_index=True)[1][1:])
        # make collids sequential
        seq_colids = np.concatenate(
            [np.repeat(i, value.shape[0]) for i, value in enumerate(grouped_array)],
            axis=0,
        )
        array_seq_colids = np.column_stack((array, seq_colids))
        # split sequential collids array by trackid and collid
        grouped_array = np.split(
            array_seq_colids,
            np.unique(array_seq_colids[:, :2], axis=0, return_index=True)[1][1:],
        )
        # generate colors for each collective event, wrap arround the color cycle
        colors = np.take(np.array(color_cylce), [i + 1 for i in np.unique(seq_colids)], mode="wrap")
        return grouped_array, colors

    def _create_noodle_plot(self, grouped_data: np.ndarray, colors: np.ndarray):
        """Plots the noodle plot."""
        fig, ax = plt.subplots()
        ax.set_xlabel("Time Point")
        ax.set_ylabel("Position")
        for dat in grouped_data:
            ax.plot(
                dat[:, 2],
                dat[:, self.projection_index],
                c=colors[int(dat[0, -1])],
            )
        return fig, ax

    def plot(self, projection_axis: str, color_cylce: list[str] = TAB20):
        """Create Noodle Plot of cell tracks, colored by collective event id.

        Arguments:
            projection_axis (str): Specify with witch coordinate the noodle
                plot should be drawn. Has to be one of the posx, posy or posz arguments
                passed in during the class instantiation process.
            color_cylce (list[str]): List of hex color values or string names
                (i.e. ['red', 'yellow']) used to color collecitve events. Cycles through list.

        Returns:
            fig, axes: Matplotlib figure and axes are returned for the noodle plot.
        """
        if projection_axis not in [self.posx, self.posy, self.posz]:
            raise ValueError(f"projection_axis has to be one of {[self.posx, self.posy, self.posz]}")
        if projection_axis == self.posx:
            self.projection_index = 3
        elif projection_axis == self.posy:
            self.projection_index = 4
        elif projection_axis == self.posz:
            self.projection_index = 5
        grpd_data, colors = self._prepare_data_noodleplot(
            self.df, color_cylce, self.colev, self.trackid, self.frame, self.posx, self.posy, self.posz
        )
        fig, axes = self._create_noodle_plot(grpd_data, colors)
        return fig, axes



"""Tools for plotting collective events."""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = '0.1.5'

from .plotting import NoodlePlot, dataPlots, plotOriginalDetrended, statsPlots

__all__ = ["plotOriginalDetrended", "dataPlots", "statsPlots", "NoodlePlot"]



import numpy as np

from arcos4py.tools import clipMeas


def test_clipping():
    data = np.linspace(1, 10, 10).astype(int)
    out = clipMeas(data).clip(0.3, 0.7)
    true_val = np.array([3.7, 3.7, 3.7, 4.0, 5.0, 6.0, 7.0, 7.3, 7.3, 7.3])
    np.testing.assert_almost_equal(out, true_val)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arcos4py', 'arcos4py.plotting', 'arcos4py.tools', 'tests']

package_data = \
{'': ['*'], 'tests': ['testdata/*']}

install_requires = \
['matplotlib>=3.3.4',
 'numpy>=1.21.5',
 'pandas>=1.3.5',
 'scikit-learn>=1.0.2',
 'scipy>=1.7.3',
 'seaborn>=0.11.2,<0.12.0']

extras_require = \
{'dev': ['tox>=3.20.1,<4.0.0',
         'virtualenv>=20.2.2,<21.0.0',
         'pip>=20.3.1,<21.0.0',
         'twine>=3.3.0,<4.0.0',
         'pre-commit>=2.12.0,<3.0.0',
         'toml>=0.10.2,<0.11.0',
         'bump2version>=1.0.1,<2.0.0'],
 'doc': ['mkdocs>=1.3.0,<2.0.0',
         'mkdocs-include-markdown-plugin>=3.3.0,<4.0.0',
         'mkdocs-material>=8.2.11,<9.0.0',
         'mkdocstrings>=0.18.0,<0.19.0',
         'mkdocs-autorefs>=0.4.1,<0.5.0',
         'mkdocstrings-python-legacy>=0.2.2,<0.3.0'],
 'test': ['black>=22.3.0,<23.0.0',
          'isort>=5.8.0,<6.0.0',
          'flake8>=3.9.2,<4.0.0',
          'flake8-docstrings>=1.6.0,<2.0.0',
          'mypy>=0.900,<0.901',
          'pytest>=6.2.4,<7.0.0',
          'pytest-cov>=2.12.0,<3.0.0',
          'PyYAML>=6.0,<7.0']}

setup_kwargs = {
    'name': 'arcos4py',
    'version': '0.1.5',
    'description': 'A python package to detect collective spatio-temporal phenomena.',
    'long_description': "# arcos4py\n\n\n[![pypi](https://img.shields.io/pypi/v/arcos4py.svg)](https://pypi.org/project/arcos4py/)\n[![conda-forge](https://img.shields.io/conda/vn/conda-forge/arcos4py)](https://anaconda.org/conda-forge/arcos4py)\n[![python](https://img.shields.io/pypi/pyversions/arcos4py.svg)](https://pypi.org/project/arcos4py/)\n[![Build Status](https://github.com/bgraedel/arcos4py/actions/workflows/dev.yml/badge.svg)](https://github.com/bgraedel/arcos4py/actions/workflows/dev.yml)\n[![codecov](https://codecov.io/gh/bgraedel/arcos4py/branch/main/graphs/badge.svg)](https://codecov.io/github/bgraedel/arcos4py)\n\n\n\nArcos4py is a python package to detect collective Spatio-temporal phenomena.\n\n* Documentation: <https://bgraedel.github.io/arcos4py>\n* GitHub: <https://github.com/bgraedel/arcos4py>\n* PyPI: <https://pypi.org/project/arcos4py/>\n* Free software: MIT\n\n\n## Features\n\nAutomated Recognition of Collective Signalling for python (arcos4py) is a python port of the R package ARCOS (https://github.com/dmattek/ARCOS) to identify collective spatial events in time-series data.\nThe software identifies collective protein activation in 2- and 3D cell cultures and can track events over time. Such collective waves have been recently identified in various biological systems and have been demonstrated to play a crucial role in the maintenance of epithelial homeostasis (Gagliardi et al., 2020, Takeuchi et al., 2020, Aikin et al., 2020),\nin the acinar morphogenesis (Ender et al., 2020), osteoblast regeneration (De Simone et al., 2021), and the coordination of collective cell migration (Aoki et al., 2017, Hino et al., 2020).\n\nDespite its focus on cell signaling, the framework can also be applied to other spatiotemporally correlated phenomena.\n\n### Todo's\n- Add additional tests for binarization and de-biasing modules.\n- Add local indicators of spatial autocorrelation (LISA) as a binarization method option.\n\nData Format\n-----------\nThe time series should be arranged in a long table format where each row defines the object's location, time, and optionally the measurement value.\n\nARCOS defines an ARCOS object on which several class methods can be used to prepare the data and calculate collective events.\nOptionally the objects used in the ARCOS class can be used individually by importing them from arcos.tools\n\nInstallation\n------------\nArcos4py can be installed from PyPI with:\n\n        pip install arcos4py\n\nNapari Plugin\n-------------\nArcos4py is also available as a Napari Plugin [arcos-gui](https://github.com/bgraedel/arcos-gui). \n[arcos-gui](https://github.com/bgraedel/arcos-gui) can simplify parameter finding and visualization.\n\n\n[![arcos_demo](https://img.youtube.com/vi/hG_z_BFcAiQ/0.jpg)](https://www.youtube.com/watch?v=hG_z_BFcAiQ)\n\n## Credits\n\n[Maciej Dobrzynski](https://github.com/dmattek) created the original ARCOS algorithm.\n\nThis package was created with [Cookiecutter](https://github.com/audreyr/cookiecutter) and the [waynerv/cookiecutter-pypackage](https://github.com/waynerv/cookiecutter-pypackage) project template.\n",
    'author': 'Benjamin Graedel',
    'author_email': 'benjamin.graedel@unibe.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bgraedel/arcos4py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



"""Module containing tools to calculate statistics of collective events.

Example:
    >>> from arcos4py.tools import calcCollevStats
    >>> test = calcCollevStats()
    >>> out = test.calculate(data = data,frame_column = "frame", collid_column = "collid")
"""

from typing import Union

import numpy as np
import pandas as pd


class calcCollevStats:
    """Class to calculate statistics of collective events."""

    def __init__(self) -> None:
        """Class to calculate statistics of collective events."""
        pass

    def _calculate_duration_size_group(self, data: np.ndarray) -> np.ndarray:
        """Calculates duration and size for the collective event in the dataframe.

        Arguments:
            data (np.ndarray): Containing a single collective event.

        Returns:
            np.ndarray: Array containing the collective event id, duration, tot_size, min_size,
                max_size, "nd_frame, first_frame_centroid and last_frame_centroid
                of the current collective event.
        """
        coll_dur = max(data[:, 0]) - min(data[:, 0]) + 1
        coll_total_size = np.unique(data[:, 1]).size
        (unique, counts) = np.unique(data[:, 0], return_counts=True)
        frequencies = np.asarray((unique, counts)).T
        coll_min_size = np.min(frequencies[:, 1])
        coll_max_size = np.max(frequencies[:, 1])
        coll_start_frame = np.min(data[:, 0])
        coll_end_frame = np.max(data[:, 0])
        if data.shape[1] > 3:
            coll_start_coord = np.mean(data[(data[:, 0] == coll_start_frame)][:, 3:], axis=0)
            coll_end_coord = np.mean(data[(data[:, 0] == coll_end_frame)][:, 3:], axis=0)
        else:
            coll_start_coord = np.nan
            coll_end_coord = np.nan
        d = np.array(
            [
                data[0, 2],
                coll_dur,
                coll_total_size,
                coll_min_size,
                coll_max_size,
                coll_start_frame,
                coll_end_frame,
                coll_start_coord,
                coll_end_coord,
            ],
            dtype=object,
        )
        return d

    def _get_collev_duration(
        self,
        data: pd.DataFrame,
        frame_column: str,
        collev_id: str,
        obj_id_column: str,
        posCol: Union[list, None],
    ) -> pd.DataFrame:
        """Applies self._calculate_duration_size_group() to every group\
        i.e. every collective event.

        Arguments:
            data (DataFrame): Containing unfiltered collective events.
            collev_id (str): Indicating the contained collective id column.
            frame_column (str): Indicating the contained frame column.
            obj_id_column (str): Indicating object id.
            posCol (list | None): Contains names of position columns. If None coordinates of
                start and end frame are not calcualted

        Returns:
            DataFrame: DataFrame containing "collid", "duration", "total_size",
                "min_size","max_size", "start_frame", "end_frame",
                "first_frame_centroid" and "last_frame_centroid"
                of all collective events.
        """
        cols = [
            collev_id,
            "duration",
            "total_size",
            "min_size",
            "max_size",
            "start_frame",
            "end_frame",
            "first_frame_centroid",
            "last_frame_centroid",
        ]
        subset = [frame_column, obj_id_column, collev_id]
        if posCol:
            subset.extend(posCol)
        # if object id cannot be converted to a float, generate unique labels with pd.factorize
        try:
            data_np = data[subset].to_numpy(dtype=np.float64)
        except ValueError:
            labels, levels = pd.factorize(data[obj_id_column])
            new_obj_id = f'{obj_id_column}_labels'
            data[new_obj_id] = labels
            subset[1] = new_obj_id
            data_np = data[subset].to_numpy(dtype=np.float64)

        data_np = data_np[~np.isnan(data_np).any(axis=1)]
        data_np_sorted = data_np[data_np[:, 2].argsort()]
        grouped_array = np.split(data_np_sorted, np.unique(data_np_sorted[:, 2], axis=0, return_index=True)[1][1:])
        # map to grouped_array
        out = map(self._calculate_duration_size_group, grouped_array)
        out_list = [i for i in out]
        df = pd.DataFrame(out_list, columns=cols)
        return df

    def calculate(
        self,
        data: pd.DataFrame,
        frame_column: str,
        collid_column: str,
        obj_id_column: str,
        posCol: Union[list, None] = None,
    ) -> pd.DataFrame:
        """Calculate statistics of collective events.

        Arguments:
            data (DataFrame): Containing collective events.
            frame_column (str): Indicating the frame column in data.
            collid_column (str): Indicating the collective event id column in data.
            obj_id_column (str): Indicating object id.
            posCol (list | None): Contains names of position columns. If None coordinates of
                start and end frame are not calcualted

        Returns:
            DataFrame: DataFrame containing "collid", "duration", "total_size",
                "min_size","max_size", "start_frame", "end_frame",
                "first_frame_centroid" and "last_frame_centroid"
                of all collective events.
        """
        if data.empty:
            return data
        colev_stats = self._get_collev_duration(data, frame_column, collid_column, obj_id_column, posCol)
        return colev_stats
